/*
  grid2cs.h
  Copyright (C) 2000 by W.C.A. Wijngaards
  Copyright (C) 2003 by TURBO J (some XML stuff added)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef GRID2CS_H
#define GRID2CS_H

class grLooksInfo;
class grSector;
class grSectorInfo;
class grLook;
class grVec;
class grTexPlaneInfo;
class grPolSet;

/// print usage and exit
void synopsis(char *programname);
/// if string differs from buf, exit()
void CheckString(char *inputbuf, char *constantstring);
/// read a look ID character or integer. If acceptspace, ' ' can be read.
int ReadLookID(FILE *in, int acceptspace=0);
/// reads a float, complains when no float is on the input.
double ReadFloat(FILE *in);
/// set global start position
void SetStart(char *secname, grVec *v);
/// get global start position
void GetStart(char *&secname, grVec *v);


/** texture information */
class grTex {
public:
  /// short name for texture, use free()
  char *name;
  /// VFS file name for texture, use free()
  char *file;
  // 0= normal, 1= procedural
  int type; 
  /// next
  grTex *next;
};

/** list of textures */
class grTexInfo {
  grTex *first;
  int amt;

public:
  /// read from file and construct;
  grTexInfo(FILE *in);

  ///
  ~grTexInfo();

  ///
  void Print(FILE *out);
  void WriteTextures(csRef<iDocumentNode> node);
};


/** looks information
*/
class grLooksInfo {
  /// nr of looks in list
  int amt;
  /// list of looks
  grLook *first;

public:
  // construct from file reading an entire Looks section.
  grLooksInfo(FILE *in);
  ///
  ~grLooksInfo();

  /// get a look based on ID
  grLook * GetLook(int id);

};


/** one 'look', determines how a grid element will look. */
class grLook {
  /// look ID
  int lookid;

  /// general properties
  double floorheight, ceilingheight;
  char *walltex;
  char *floortex;
  char *ceiltex;

  /// light properties
  /// if true it has a light
  int has_light; 
  double radius;
  double r,g,b;
  double lightheight;

  /// start position?
  int is_start;

  /// next look in list
  grLook *next;
public:
  /** 
   * skipping 'Look' is done by caller of constructor.
   */
  grLook(FILE *in);
  ///
  ~grLook();

  /// Get Look id;
  int GetLookID() const {return lookid;}

  /// must override, really
  double GetFloorHeight() {return floorheight;}
  double GetCeilingHeight() {return ceilingheight;}

  /// Get wall texture
  char *GetWallTexture() const { return walltex; }
  /// get next look
  grLook *GetNext() const {return next;}
  /// set next look
  void SetNext(grLook *n) {next = n;}

  /// create polygons for this look at a gridpoint in a sector
  void Polygonize(grSector *sec, int x, int y);
  /** add walls. (to neighbor for portals to leave open)
   * dir is to the right. shift is shift in gridpoints.
   */
  void addwall(grSector *sec, int x, int y, int neighbor_x, int neighbor_y,
    int dir_x, int dir_y, float shift_x, float shift_y);

  /// print special info to be inside the SECTOR statement
  void PrintInSector(int x, int y, FILE *out, grSector *sec);
  void WriteLook(int x, int y, csRef<iDocumentNode> node, grSector* sec);
};


/** Map
 */
class grMap {
  /// size of map (xwidth, yheight)
  int width, height;
  /// width*height
  int size;
  /// lookids,  ' ' is unreachable area
  int *map;

  /// looks to use for this map
  grLooksInfo *lookinfo;
public:
  /// make
  grMap(FILE *in);
  ///
  ~grMap();

  /// get width
  int GetWidth() const {return width;}
  /// get height
  int GetHeight() const {return height;}
  /// get size (width * height)
  int GetSize() const {return size;}

  /// Get a look by x,y. Even works without looksinfo set.
  int GetLookID(int x, int y) const {return map[y*width+x];}
  /// Get a grLook by x,y needs looksinfo set.
  grLook *GetLook(int x, int y);
  
  /// Check if a x,y is reachable. x,y may be out of range(returns unreachable)
  /// works without lookinfo set. Checks for lookid ' '.
  int IsReachable(int x, int y);
  /// set looksinfo, so other information functions will work.
  void SetLooksInfo(grLooksInfo *inf) {lookinfo = inf;}
  /// get floor height of a spot
  double GetFloorHeight(int x, int y);
  /// get ceiling height of a spot
  double GetCeilingHeight(int x, int y);

  /** convert a grid x and grid y to a 3d coordinate, given a height.
   *  in gridpoint (0,0), the value (0,0) is the center point,
   *  with x=0, z=0 in 3D, 
   *  (0.5,0.5) is the bottom-right corner, with x=0.5, z=-0.5 in 3D.
   *  make a copy of result if you plan to keep it.
   */
  grVec *Get3dPos(float x, float y, float height);

  /// print map for debug
  void PrintMap();
};



/** one sector */
class grSector {
public:
  /// sectorization to which the sector belongs.
  grSectorInfo *info;
  /// unique number for sectors
  int sector_num;
  /// x, y, width, height (examine map for contents of sector)
  int x, y, w, h;
  /// the polygons in this sector
  grPolSet *pols;

  /// next sector in list
  grSector *next;

public:
  /// create this
  grSector(grSectorInfo *parent);
  ///
  ~grSector();
  /// compute polygons and portals for this sector
  void Polygonize();
  /** create portals by scanning the gridline starting at start_x,y,
  * running in the dir_x,y direction, with a length of todo_dist.
  * Note that these points must not be part of this sector.
  * Note also that the dir_x,y must be 'towards the right' when viewed
  *  from the center of this sector.
  * All points of the polygon are shiftx,y-ed. So that you can make
  *  polygons that do not go through the middle.
  * distance == the number of grid points.
  */
  void create_portals( int start_x, int start_y, int dir_x, int dir_y, 
    int todo_dist, float shift_x, float shift_y);

  /// Get the unique name for the sector.
  char *GetName();

  /// Get the polygon set
  grPolSet *GetPolSet() const {return pols;}
  /// Get The SectorInfo of this sector
  grSectorInfo *GetSectorInfo() const {return info;}

  /// print sector to file
  void Print(FILE *out);
  void WriteSector(csRef<iDocumentNode> node);

};

/** sector info */
class grSectorInfo
{
  /// linked list of sectors found.
  grSector *first;
  /// number of sectors in info.
  int num;
  /// the map sectorized by this sectorinfo
  grMap *map;
  /// the texture planes collection
  grTexPlaneInfo *texplanes;

  /// flags, 0x1 means sectorized area.
  int *flags;
  /// the sector to which a grid point belongs
  grSector **sector;

  /// statistic: nr of polygons
  int nr_poly;
  /// statistic: nr of portals
  int nr_portal;
  /// statistic: nr of vertices
  int nr_vert;
public:
  ///
  grSectorInfo();
  ///
  ~grSectorInfo();

  /// divide the given map into sectors.
  void Sectorize(grMap *map);

  /// Set Map (creates flags, sector lookup)
  void SetMap(grMap *m);
  /// is sectorized? 
  int GetSectorized(int x, int y) const {return flags[y*map->GetWidth()+x]&0x1;}
  /// set sectorized
  void SetSectorized(int x, int y) {flags[y*map->GetWidth()+x] |= 0x1;}
  /// set sectorized for an area
  void SetSectorized(grSector *sec, int x, int y, int w, int h);

  /// get a sector that covers x,y
  grSector *GetSector(int x, int y) const {return sector[y*map->GetWidth()+x];}
  /** find a stretch of points that have the same sector. Unreachable is
    also a type of sector here, but false is returned when that is the case.
    Also the distance is returned (from start in dir)*/
  int SameSector(int start_x, int start_y, int dir_x, int dir_y, 
    int todo_dist, int& found_dist);

  /// find nonsectorized reachable area in a map. returns spot and true.
  int findnewarea(int &x, int &y);
  /// check if area designated has given heights and is reachable nonsectorized.
  int checkarea(int x, int y, int w, int h, double floor, double ceiling);
  /// grow an area from start spot to biggest rectangle
  void growarea(int x, int y, int &w, int &h);

  /// compute the polygons for all sectors.
  void Polygonize(grTexPlaneInfo *tplanes);
  /// Get Map (only after a setmap)
  grMap *GetMap() const { return map; }
  /// Get textureplaneinfo, after or during a polygonize
  grTexPlaneInfo *GetTexPlaneInfo() const { return texplanes; }

  /// Get nr of sectors in this info
  int GetNumberSectors() const {return num;}
  
  ///
  void Print(FILE *out);
  /// debug print the sectorization on output, useing ' ' and '0'+sec_num
  void PrintSectors();
  void WriteSectors(csRef<iDocumentNode> node);
};








class grVec {
public:
  /// value
  float x, y, z;

  grVec() {x=0.0; y=0.0; z=0.0;}
  grVec(float x1, float y1, float z1) {x=x1; y=y1; z=z1;}
  void Set(float x1, float y1, float z1) {x=x1; y=y1; z=z1;}
};


/** texture mapping plane information */
class grTexPlane {
public:
  /// texture name
  char *texture_name;
  /// plane name (unique)
  char *plane_name;
  /** points on the plane, orig, v1, and v2. (ORIG, FIRST, SECOND)
   Thus v1 is to the right of orig, and v2 is down from orig. */
  grVec *orig, *v1, *v2;

  /// next plane
  grTexPlane *next;
  void WritePlane(csRef<iDocumentNode> node);
};


/** list of texture mapping planes */
class grTexPlaneInfo {
  grTexPlane *first;
  int amt;
public:
  grTexPlaneInfo();
  ~grTexPlaneInfo();

  /// returns plane name, checks for duplicates.
  grTexPlane* AddPlane(char *tex, grVec *orig, grVec *v1, grVec *v2);
  /// Add a plane at given point and dir_x,y pointing to the right on a map
  grTexPlane* AddPlane(char *tex, grVec *pt, int dir_x, int dir_y);
  /// Print Planes
  void Print(FILE *out);
  
};

/** polygon information */
class grPol {
public:
  /// nr of vertices
  int amt;
  /// vertices (indices into vertex array of parent)
  int *verts;
  /// texture name
  char *texture;
  /// texture mapping plane name
  grTexPlane *plane;
  /// if non-NULL this is a portal to the named sector.
  char *portal_dest;

  /// next polygon in list
  grPol *next;
};
/** a set of polygons sharing vertices */
class grPolSet {
  /// amount of vertices
  int nr_v;
  /// amount of polygons
  int nr_p;
  /// size of arrays
  int max_v;
  /// polygons (linked list)
  grPol *first;
  /// vertices (array)
  grVec *verts;
  /// (statistic only) number of portals among the polygons
  int number_portals;
public:
  /// init to empty
  grPolSet();
  ///
  ~grPolSet();

  /// Add new vertice to the set, index is returned.
  int AddVertice(float x, float y, float z);
  /// Add a new certice, convenience version
  int AddVertice(grVec *vec);

  /// Add a new square polygon to the set, using the given indices.
  void AddPolygon(int i1, int i2, int i3, int i4,
    char *texture, grTexPlane *texplane, char *portal_dest = 0);

  /// Get the number of polygons
  int GetNumberPolygons() const {return nr_p;}
  /// Get the number of portals (among the total polygon amount)
  int GetNumberPortals() const {return number_portals;}
  /// Get the number of vertices
  int GetNumberVertices() const {return nr_v;}


  /// Print the polygonset
  void Print(FILE *out);
  void WritePolSET(csRef<iDocumentNode> node);
};


#endif // GRID2CS_H
